import {API_URL} from '@env';
import {
  ApolloClient,
  createHttpLink,
  DefaultOptions,
  from,
  fromPromise,
  InMemoryCache,
  split,
} from '@apollo/client';
import {setContext} from '@apollo/client/link/context';
import {onError} from '@apollo/client/link/error';
import {GraphQLWsLink} from '@apollo/client/link/subscriptions';
import {getMainDefinition} from '@apollo/client/utilities';
import {createClient} from 'graphql-ws';
import {
  getStorageData,
  getStorageObjectData,
  setStorageObjectData,
  StorageKey,
} from '../utils/storage';

type ResponseRefreshToken = {
  accessToken: string;
  refreshToken: string;
};

export const useClient = () => {
  const refreshToken = async (): Promise<{refreshToken: string; accessToken: string}> => {
    try {
      const userStorage = await getStorageObjectData<{refreshToken: string}>(StorageKey.DATAUSER);

      const request = await fetch(API_URL, {
        method: 'POST',
        headers: {
          'Content-Type': 'application/json',
        },
        body: JSON.stringify({
          query: `
              query RefreshToken($refreshToken: String!) {
                refreshToken(refreshToken: {refreshToken: $refreshToken}) {
                  accessToken
                  refreshToken
                }
              }
            `,
          variables: {
            refreshToken: userStorage?.refreshToken ?? '',
          },
        }),
      });

      const response: ResponseRefreshToken = await request.json();

      return response;
    } catch (error) {
      console.error(error);
      return {
        refreshToken: '',
        accessToken: '',
      };
    }
  };

  const errorLink = onError(({graphQLErrors, networkError, operation, forward}) => {
    if (graphQLErrors)
      graphQLErrors.forEach(({message, locations, path, extensions, ...other}) => {
        switch (extensions.code) {
          case 'invalid-jwt':
            return fromPromise(
              refreshToken()
                .then(responsePromise => {
                  getStorageObjectData<{name: string}>(StorageKey.DATAUSER).then(user => {
                    setStorageObjectData(StorageKey.DATAUSER, {
                      ...user,
                      refreshToken: responsePromise.refreshToken,
                      token: responsePromise.accessToken,
                    });
                  });
                  const oldHeaders = operation.getContext().headers;
                  // modify the operation context with a new token
                  operation.setContext({
                    headers: {
                      ...oldHeaders,
                      Authorization: `Bearer ${responsePromise.accessToken}`,
                    },
                  });

                  // retry the request, returning the new observable
                  return forward(operation);
                })
                .catch(error => {
                  console.error(error);
                }),
            );
        }
        console.error(
          `[GraphQL error]: Message: ${message}, Location: ${locations}, Path: ${path}, other: ${JSON.stringify(
            other,
          )} `,
        );
      });
    if (networkError) {
      console.error(`[Network error]: ${networkError}`);
    }
  });

  const authLink = setContext(async setter => {
    const user = await getStorageObjectData<{token: string; role: string}>(StorageKey.DATAUSER);
    console.log(`Bearer ${user?.token}`);

    return {
      ...setter,
      headers: {
        ...(user?.token && {
          Authorization: `Bearer ${user.token}`,
          'x-hasura-role': user?.role || '',
        }),
      },
    };
  });

  const httpLink = createHttpLink({
    uri: API_URL,
  });

  const wsLink = new GraphQLWsLink(
    createClient({
      url: `ws://${API_URL}`,
    }),
  );

  const appLink = from([errorLink, httpLink]);

  const splitLink = split(
    ({query}) => {
      const definition = getMainDefinition(query);
      return definition.kind === 'OperationDefinition' && definition.operation === 'subscription';
    },
    wsLink!,
    authLink.concat(appLink),
  );

  const defaultOptions: DefaultOptions = {
    watchQuery: {
      fetchPolicy: 'no-cache',
      errorPolicy: 'all',
    },
    query: {
      fetchPolicy: 'no-cache',
      errorPolicy: 'all',
    },
    mutate: {
      fetchPolicy: 'no-cache',
      errorPolicy: 'all',
    },
  };
  return {
    client: new ApolloClient({
      link: splitLink,
      cache: new InMemoryCache({addTypename: false}),
      defaultOptions: defaultOptions,
    }),
  };
};
